<?php
/**************************************************************
 * Subset of Sass for PHP
 * (c) 2009, artoodetoo.
 *
 * Sass homepage in http://sass-lang.com/
 **************************************************************/

class SassFilter
{
	const TAB = 4;
	// Charset for lexer
	const ID_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_';
	// Sass string kind
	const VAR_DEF	= 1;
	const MIXIN_DEF	= 2;
	const MIXIN_USE	= 3;
	const RULE_DEF	= 4;
	const ATTR_PREF	= 5;
	const ATTR_DEF	= 6;
	const ATTR_CALC	= 7;
	const COMMAND	= 8;
	// Output style
	const OS_EXPANDED	= 0;
	const OS_NESTED		= 1;
	const OS_COMPACT	= 2;
	const OS_COMPRESSED	= 3;

	private static $_callback;
	private static $_head;
	private static $_level;
	private static $_stack;
	private static $_names;
	private static $_variables;
	private static $_mixins;
	private static $_rules;

	private static function _cleanStr($s)
	{
		if(($p = strpos($s, '//')) !== FALSE)
			$s = substr($s, 0, $p);
		$s = rtrim($s);
		while (($p = strpos($s, "\t")) !== FALSE)
			$s = substr_replace($s, str_repeat(' ', self::TAB - $p % self::TAB), $p, 1);
		return $s;
	}

	private static function &_blockByOffset($offs)
	{
		$ret = NULL;
		foreach (self::$_stack as $o => $item)
		{
			if ($o >= $offs)
				break;
			if ($item[1] == self::MIXIN_DEF)
				$ret =& self::$_mixins[$item[2]];
			else if ($item[1] == self::RULE_DEF)
				$ret =& self::$_rules[$item[0]];
		}
		return $ret;
	}

	private static function _nestedAttrNames($offs)
	{
		$ret = '';
		foreach (self::$_stack as $o => $item)
		{
			if ($o > $offs) break;
			if (self::ATTR_PREF <= $item[1] && $item[1] <= self::ATTR_CALC)
				$ret = ($ret != '' ? $ret.'-' : '') . $item[2];
		}
		return $ret;
	}

	private static function _nestedSelectors($offs)
	{
		$childSelector = self::$_names[self::$_stack[$offs][0]]['name'];
		$parentSelector = NULL;

		foreach (self::$_stack as $o => $item)
		{
			if ($o >= $offs) break;
			if ($item[1] == self::RULE_DEF)
				$parentSelector = self::$_names[$item[0]]['name'];
		}
		if (!is_null($parentSelector))
		{
			$parentArray = explode(',', $parentSelector);
			$childArray = explode(',', $childSelector);
			foreach ($parentArray as $parent)
			{
				foreach ($childArray as &$child)
				{
					$child = (($ampPos = strpos($child, '&')) !== FALSE) ?
						substr_replace($child, $parent, $ampPos, 1) :
						$child = $parent . ' ' . $child;
				}
			}
			$childSelector = str_replace('  ', ' ', implode(', ', $childArray));
		}

		return $childSelector;
	}

	private static function _parseStr($s)
	{
		$value = NULL;
		$s = ltrim($s);
		switch ($s{0}) {
		case '@':
			$name = substr($s, 1, $l = strspn($s, self::ID_CHARS, 1));
			$value = ltrim(substr($s, 1 + $l));
			$kind = self::COMMAND;
			break;
		case '!':
			$name = substr($s, 1, strspn($s, self::ID_CHARS, 1));
			$value = ltrim(substr($s, 1 + strpos($s, '=')));
			$kind = self::VAR_DEF;
			break;
		case '=':
			$name = substr($s, 1, strspn($s, self::ID_CHARS, 1));
			$kind = self::MIXIN_DEF;
			break;
		case '+':
			$name = substr($s, 1, strspn($s, self::ID_CHARS, 1));
			$value = self::$_mixins[$name];
			$kind = self::MIXIN_USE;
			break;
		case ':':
			$name = substr($s, 1, $l = strspn($s, self::ID_CHARS, 1));
			$value = ltrim(substr($s, 1 + $l));
			if ($value == '')
				$kind = self::ATTR_PREF;
			else if ($value{0} == '=')
			{
				$value = ltrim(substr($value, 1));
				$kind = self::ATTR_CALC;
			}
			else
				$kind = self::ATTR_DEF;
			break;
		default:
			$name = $s;
			$kind = self::RULE_DEF;
		}
		return array($kind, $name, $value);
	}

	private static function _processStr($s)
	{
		$key = count(self::$_names);
		$offs = strspn($s, ' ');
		list($kind, $name, $value) = self::_parseStr($s);
		self::$_names[$key] = array('name' => $name, 'kind' => $kind, 'offs' => $offs);

		$stack =& self::$_stack;
		while (end($stack) && key($stack) >= $offs)
			unset($stack[key($stack)]);
		$stack[$offs] = array($key, $kind, $name);

		switch ($kind) {
		case self::VAR_DEF:
			self::$_variables[$name] = SassExpression::calc($value, self::$_variables);;
			break;
		case self::MIXIN_DEF:
			self::$_mixins[$name] = array();
			break;
		case self::MIXIN_USE:
			$target =& self::_blockByOffset($offs);
			foreach ($value as $line)
				$target[] = $line;
			break;
		case self::RULE_DEF:
			self::$_names[$key]['name'] = self::_nestedSelectors($offs);
			self::$_rules[$key] = array();
			break;
		case self::ATTR_PREF:
			break;
		case self::ATTR_DEF:
			$target =& self::_blockByOffset($offs);
			$finalName = self::_nestedAttrNames($offs);
			$target[] = array($finalName, $value);
			break;
		case self::ATTR_CALC:
			$target =& self::_blockByOffset($offs);
			$finalName = self::_nestedAttrNames($offs);
			$target[] = array($finalName, SassExpression::calc($value, self::$_variables));
			break;
		case self::COMMAND:
			if (!is_null(self::$_callback))
				call_user_func(self::$_callback, $s);
			break;
		}
	}

	public static function input($sass, $callback = NULL)
	{
		if (++self::$_level == 1)
		{
			self::$_callback = $callback;
			self::$_stack = self::$_names = self::$_variables = self::$_mixins = self::$_rules = self::$_head = array();
		}

		$lines = array_filter(array_map(array('self', '_cleanStr'), explode("\n", $sass)), 'strlen');
		array_walk($lines, array('self', '_processStr'));

		--self::$_level;
	}

	public static function output($outStyle = SassFilter::OS_NESTED)
	{
		$ret = empty(self::$_head) ? '' : implode("\n", self::$_head)."\n\n";
		switch ($outStyle) {
		case self::OS_EXPANDED:
			foreach (self::$_rules as $key => $attrs)
			{
				if (empty($attrs)) continue;
				$ret .= self::$_names[$key]['name'] . ' { ';
				foreach ($attrs as $attr)
					$ret .= "\n\t" . $attr[0] . ': ' . $attr[1] . ';';
				$ret .= "\n}\n\n";
			}
			break;
		case self::OS_NESTED:
			foreach (self::$_rules as $key => $attrs)
			{
				$tab = str_repeat(' ', self::$_names[$key]['offs']);
				$ret .= $tab . self::$_names[$key]['name'] . ' { ';
				foreach ($attrs as $attr)
					$ret .= "\n" . $tab . '  ' . $attr[0] . ': ' . $attr[1] . ';';
				$ret .= "}\n";
			}
			break;
		case self::OS_COMPACT:
			foreach (self::$_rules as $key => $attrs)
			{
				if (empty($attrs)) continue;
				$ret .= self::$_names[$key]['name'] . ' { ';
				foreach ($attrs as $attr)
					$ret .= "" . $attr[0] . ': ' . $attr[1] . '; ';
				$ret .= "}\n";
			}
			break;
		case self::OS_COMPRESSED:
			foreach (self::$_rules as $key => $attrs)
			{
				if (empty($attrs)) continue;
				$ret .= self::$_names[$key]['name'] . '{';
				foreach ($attrs as $attr)
					$ret .= "" . $attr[0] . ':' . $attr[1] . ';';
				$ret .= "}";
			}
			break;
		}
		return $ret;
	}

	public static function pipe($sass, $outStyle = SassFilter::OS_NESTED)
	{
		self::input($sass);
		return self::output($outStyle);
	}

	public static function addHead($s)
	{
		self::$_head[] = $s;
	}
}


class SassExpression
{
	/*
	const STR_PATTERN1 = "'.*?'";
	const STR_PATTERN2 = '".*?"';
	const VAR_PATTERN = '\![\w\-_]+';
	*/
	// Value & Operation stacks
	private static $_E;
	private static $_T;
	// array_walk cannot return value, so put it here
	private static $_result;
	// Variable list as (string Name => string Value)
	private static $_vars;
	// Operation ID is a key of $_operations array
	const OP_ADD = 2;
	const OP_SUB = 3;
	const OP_MUL = 4;
	const OP_DIV = 5;
	private static $_operations = array(
		   NULL, '(', '+', '-', '*', '/', ')');
	// Rules defined as [stackTopID][nextInputID] => rule
	private static $_rules = array(
		array(6,   1,   1,   1,   1,   1,   5),
		array(5,   1,   1,   1,   1,   1,   3),
		array(4,   1,   2,   2,   1,   1,   4),
		array(4,   1,   2,   2,   1,   1,   4),
		array(4,   1,   4,   4,   2,   2,   4),
		array(4,   1,   4,   4,   2,   2,   4));
	// Operand type
	const OT_NUMBER = 1;
	const OT_SIZE   = 2;
	const OT_COLOR  = 3;
	const OT_STRING = 4;
	// Pre-defined colors
	private static $_colors = array(
		'black'   => array(0x00, 0x00, 0x00),
		'gray'    => array(0x80, 0x80, 0x80),
		'silver'  => array(0xc0, 0xc0, 0xc0),
		'white'   => array(0xff, 0xff, 0xff),
		'red'     => array(0xff, 0x00, 0x00),
		'fuchsia' => array(0xff, 0x00, 0xff),
		'maroon'  => array(0x80, 0x00, 0x00),
		'purple'  => array(0x80, 0x00, 0x80),
		'navy'    => array(0x00, 0x00, 0x80),
		'blue'    => array(0x00, 0x00, 0xff),
		'aqua'    => array(0x00, 0xff, 0xff),
		'green'   => array(0x00, 0x80, 0x00),
		'lime'    => array(0x00, 0xff, 0x00),
		'teal'    => array(0x00, 0x80, 0x80),
		'yellow'  => array(0xff, 0xff, 0x00),
		'olive'   => array(0x80, 0x80, 0x00));
	// Size dimensions
	private static $_dimensions = array(
		       'in', 'cm', 'mm', 'pt', 'em', 'wm', 'px', '%');
	private static $_dimRatio = array(
		array(    1, 2.54, 25.5,  72    ),
		array(0.394, 1,    10,    28.346),
		array(0.039, 0.1,   1,     2.835),
		array(0.014, 0.035, 0.353, 1    ));

	//
	// Process next token. It is callback for array_reduce()
	//   integer topT  - operation in top of stack
	//   string  token - next input token
	//
	private static function _token($token)
	{
		// Look up for operation ID
		if (($curOp = array_search($token, self::$_operations)) === FALSE)
		{
			// Token is value
			array_push(self::$_E, self::_strToValue($token));
		}
		else
		{
			// Operation stack top
			$topT = end(self::$_T);
			// Token is operation
			while (TRUE) // loop for rule #4 only!
			{
				switch ($f = self::$_rules[$topT][$curOp]) {
				case 1:
					array_push(self::$_T, $curOp);
					break;
				case 2:
				case 4:
					$b  = array_pop(self::$_E);
					$op = array_pop(self::$_T);
					$a  = array_pop(self::$_E);
					if (is_null($a) || is_null($b))
						throw new Exception('missing operand');
					array_push(self::$_E, self::_func($a, $op, $b));
					if ($f == 4)
					{
						$topT = end(self::$_T);
						continue 2;
					}
					array_push(self::$_T, $curOp);
					break;
				case 3:
					array_pop(self::$_T);
					break;
				case 5:
					throw new Exception('unclosed brackets');
				case 6:
					if (count(self::$_E) != 1)
						throw new Exception('missing operator');
					else
						self::$_result = self::_valueToStr(self::$_E[0]);
				}

				break; // all cases except rule #4
			} // end of loop
		}
	}

	private static function _valueToStr($arr)
	{
		list($x, $type, $dim) = $arr;
		switch ($type) {
		case self::OT_NUMBER: return $x;
		case self::OT_SIZE:   return $x . self::$_dimensions[$dim];
		case self::OT_COLOR:  return sprintf('#%02x%02x%02x', $x[0], $x[1], $x[2]);
		case self::OT_STRING: return '"' . $x . '"';
		}
	}

	private static function _strToValue($x)
	{
		$type = self::OT_NUMBER;
		$dim = NULL;

		if ($x{0} == '!')
		{
			$x = substr($x, 1);
			if (!isset(self::$_vars[$x]))
				throw new Exception('undefined variable "'.$x.'"');
			$x = self::$_vars[$x];
		}

		if ($x{0} == '"' || $x{0} == '\'')
		{
			$type = self::OT_STRING;
			$x = trim($x, $x{0});
		}
		else if (isset(self::$_colors[$x]))
		{
			$type = self::OT_COLOR;
			$x = self::$_colors[$x];
		}
		else if ($x{0} == '#')
		{
			$type = self::OT_COLOR;
			$x = substr($x, 1);
			$len = strlen($x);
			if (!ctype_xdigit($x) || ($len != 3 && $len != 6))
				throw new Exception('wrong color constant #'.$x);
			if ($len == 3)
				$x = array(hexdec($x{0}.$x{0}), hexdec($x{1}.$x{1}), hexdec($x{2}.$x{2}));
			else
				$x = array(hexdec($x{0}.$x{1}), hexdec($x{2}.$x{3}), hexdec($x{4}.$x{5}));
		}
		else if (substr($x, -1) == ($d = '%') || in_array($d = substr($x, -2), self::$_dimensions))
		{
			$type = self::OT_SIZE;
			$dim = array_search($d, self::$_dimensions);
			$x = substr($x, 0, -strlen($d));
		}

		return array($x, $type, $dim);
	}

	//
	// Execute simple operation
	//  mixed   a, b - operands (PHP will interret strings as numbers)
	//  integer op   - operation ID
	//  returns numeric operation result
	//
	private static function _funcNumber($a, $op, $b)
	{
			switch ($op) {
			case self::OP_ADD: return $a + $b;
			case self::OP_SUB: return $a - $b;
			case self::OP_MUL: return $a * $b;
			case self::OP_DIV: return $b == 0 ? 0 : $a / $b;
			}
	}

	private static function _func($a, $op, $b)
	{
		list($a, $aType, $aDim) = $a;
		list($b, $bType, $bDim) = $b;

		$resultType = max($aType, $bType);
		if (min($aType, $bType) == self::OT_SIZE && $resultType == self::OT_COLOR)
			throw new Exception('color and size are not compatible');

		switch ($resultType) {
		case self::OT_NUMBER:
			return array(self::_funcNumber($a, $op, $b), $resultType, NULL);

		case self::OT_SIZE:
			if (is_null($aDim)) $aDim = $bDim;
			if (is_null($bDim)) $bDim = $aDim;
			if ($aDim != $bDim)
			{
				if (!isset(self::$_dimRatio[$aDim][$bDim]))
					throw new Exception('"'.self::$_dimensions[$aDim].'" and "'.self::$_dimensions[$bDim].'" are not compatible');
				$b *= self::$_dimRatio[$bDim][$aDim];
			}
			return array(self::_funcNumber($a, $op, $b), $resultType, $aDim);

		case self::OT_COLOR:
			if ($aType == self::OT_NUMBER) $a = array($a, $a, $a);
			if ($bType == self::OT_NUMBER) $b = array($b, $b, $b);
			foreach ($a as $key => $dummy)
				$a[$key] = max(0, min(255, self::_funcNumber($a[$key], $op, $b[$key])));
			return array($a, $resultType, NULL);

		case self::OT_STRING:
			if ($op == self::OP_ADD)
				return array($a . $b, $resultType, NULL);
			break;
		}
		throw new Exception('"'.self::$_operations[$op].'" is wrong operation for these arguments');
	}

	//
	// Process Sass expression
	//  string s - expression
	//  array  v - variables as (name => value)
	//
	public static function calc($s, $v = NULL)
	{
		/*static $pattern;*/

		self::$_vars = $v;
		self::$_E = array();
		self::$_T = array(0); // put bottom of stack marker

		// lexer
		/*if (is_null($pattern))
		{
			$pattern = array_slice(self::$_operations, 1);
			$pattern = array_map('preg_quote', $pattern);
			array_unshift($pattern, self::STR_PATTERN1, self::STR_PATTERN2, self::VAR_PATTERN);
			$pattern = '#\s*(' . implode('|', $pattern) . ')\s*#';
			print_r($pattern);
		}
		$split = preg_split($pattern, $s, -1, PREG_SPLIT_DELIM_CAPTURE|PREG_SPLIT_NO_EMPTY);*/
		$split = preg_split('#\s*(\'.*?\'|".*?"|\![\w\-_]+|\+|-|\*|/|\(|\))\s*#', $s, -1, PREG_SPLIT_DELIM_CAPTURE|PREG_SPLIT_NO_EMPTY);
		$split[] = NULL; // put end of tokens marker

		try {
			array_walk($split, array('self', '_token'));
			return self::$_result;
		} catch (Exception $e) {
			return 'ERROR: ' . 'line ' . $e->getLine() . ' ' . $e->getMessage();
		}
	}
}
